#include "common.h"
#include "const.h"
#include <iostream>
#include <fstream>
#include <cmath>
#include <time.h>

using namespace std;

#include <cstring>
#include <cstdlib>

double Get_distance(LOCATION , LOCATION );

/*
/*===================================================================
FUNCTION: UE::Channel_Update_MIMO( void )



===================================================================*/
void MS::Channel_Update_MIMO(int ms_idx)
{
	_self_idx = ms_idx;

	/*===================================================================
	FUNCTION : Declare_Ch_matrix()
	OUTPUT   : H_m, HI_m(interference channel)
	===================================================================*/
	if (CH_CAL == 1)
	{
		Declare_ch_matrix();
	}
	/*===================================================================
	FUNCTION : Fourier_Transform_of_Channel(_self_idx)
	OUTPUT   : H_m, HI_m(interference channel)
	===================================================================*/
	Fourier_Transform_of_Channel(_self_idx);

	/*===================================================================
	FUNCTION : Quantization_of_Ch()
	INPUT    : H_m
	OUTPUT   : PMI(cw_idx )
	===================================================================*/
	Quantization_of_Ch();

	/*===================================================================
	FUNCTION : CQI_Update()
	INPUT    : H_m, UE_PATH_LOSS, UE_ICI
	OUTPUT   : CQI(esinr)
	===================================================================*/
	CQI_Update();


	

}

void MS::Quantization_of_Ch(void)
{
	if (CH_CAL == 1) //// run time calc
	{
		for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
		{
			
			if (false ) //(NUM_TX == 4)   /// PMI Decision
			{
				Matrix4cd Ohm;

				double min = 500;

				for (int cw_idx = 0; cw_idx < 16; cw_idx++)
				{

					Ohm = WW[cw_idx] * (H_m[rbs_idx].adjoint() * H_m[rbs_idx]) * WW[cw_idx].adjoint(); //Ohm = W HH H WH // 4x4 matrix

					Matrix4cd UppOhm;
					Matrix4cd LowOhm;

					UppOhm = Ohm.triangularView<StrictlyUpper>(); //upper triangle matrix except diagonal 
					LowOhm = Ohm.triangularView<StrictlyLower>(); //lower triangle matrix except diagonal



					double temp = UppOhm.norm() + LowOhm.norm(); // minimizes off diagonal entries of Ohm

					if (temp < min)
					{
						min = temp;
						PMI[rbs_idx][t%cqi_history_length] = cw_idx;
					}
				}
			}
			else if (false)  //// old code PMI decision
			{
				JacobiSVD <MatrixXcd> svd(H_m[rbs_idx], ComputeFullU | ComputeFullV);    ////// U = 2x2, V = 4x4

				/*
				cout << "Its singular values are:" << endl << svd.singularValues() << endl;
				cout << "Its left singular vectors are the columns of the thin U matrix:" << endl << svd.matrixU() << endl;
				cout << "Its right singular vectors are the columns of the thin V matrix:" << endl << svd.matrixV() << endl;

				getchar();
				*/

				MatrixXcd H_eq;

				H_eq = (svd.matrixV().col(0)).adjoint();  ///////// 1x4 vector

				//cout << "H_eq = " << H_eq << endl;

				double max = -500;

				for (int cw_idx = 0; cw_idx < 64; cw_idx++)
				{
					double temp = (H_eq * code_word[cw_idx].conjugate()).norm();

					if (temp > max)
					{
						max = temp;
						PMI[rbs_idx][t%cqi_history_length] = cw_idx;
					}
				}

				//cout << "Precalc_PMI = " << PMI_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;
				//getchar();

			}

			else /// Random PMI 
			{
				PMI[rbs_idx][t%cqi_history_length] = rand() % 16;
			}
		}
	}
}





void MS::CQI_Update(void)
{

	if (CH_CAL == 1)  //// run-time calculate
	{
		//if (NUM_TX != 4)   // No Use PMI
		if (Scheduling_Type == 0)  // round robin scheduling
		{
			for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
			{
				double sum_of_ch_coef = 0.;
				//complex<double> s(1, 2);
				for (int r_idx = 0; r_idx < NUM_RX; r_idx++) // Received    Channel
				{
					for (int t_idx = 0; t_idx < NUM_TX; t_idx++) // Transmitted Channel
					{
						sum_of_ch_coef += pow(abs(H_m[rb_idx](r_idx, t_idx)), 2.);
					}
				}

				double geometry = link[_self_idx].str_signal - link[_self_idx].interference; // (dB scale) // == (link[_self_idx].UE_PATH_LOSS / link[_self_idx].UE_ICI) (linear scale)

				double linear_signal = dBm2linear(link[_self_idx].str_signal);
				double linear_interference = dBm2linear(link[_self_idx].interference);

				linear_interference = linear_interference + noise;
				
				CQI[rb_idx][abs(t) % cqi_history_length] = sum_of_ch_coef * (linear_signal / linear_interference); /// 

				//cout << "CQI = " << CQI[rb_idx][abs(t) % cqi_history_length] << endl;
				//getchar();
			}

		}

		else if (Scheduling_Type == 1)  //MU MIMO scheduling..
		{
			double noise = pow(10, (-174.0 / 10.0)) * (bandwidth);  // bandwidth = Hz
			MatrixXcd tempRI, tempRH, tempRHr, tempM, tempIRC, temph, eye, signal, interferencePlusNoise;
			tempRI = MatrixXcd::Zero(NUM_RX, NUM_RX);
			tempRHr = MatrixXcd::Zero(NUM_RX, NUM_RX);
			tempRH = MatrixXcd::Zero(NUM_TX, NUM_TX);
			eye = MatrixXcd::Identity(NUM_RX, NUM_RX);

			// For SVD
			MatrixXcd tempV;


			for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
			{
				tempRHr = tempRHr + (H_m[rb_idx] * H_m[rb_idx].transpose()) / num_rbs_in_1ms;   /// tempRHr = num_RX x num_RX
				tempRH  = tempRH +  (H_m[rb_idx].transpose() * H_m[rb_idx]) / num_rbs_in_1ms;   /// tempRH  = num_TX x num_TX
			}

			JacobiSVD <MatrixXcd> svd(tempRH, ComputeFullU | ComputeFullV);    ////// U = RX x RX, V = TX x TX

			/*
			cout << "Its singular values are:" << endl << svd.singularValues() << endl;
			cout << "Its left singular vectors are the columns of the thin U matrix:" << endl << svd.matrixU() << endl;
			cout << "Its right singular vectors are the columns of the thin V matrix:" << endl << svd.matrixV() << endl;
			Vector3f rhs(1, 0, 0);
			cout << "Now consider this rhs vector:" << endl << rhs << endl;
			cout << "A least-squares solution of m*x = rhs is:" << endl << svd.solve(rhs) << endl;
			*/


			tempV = svd.matrixV();
			digitalPrecodingMatrix = tempV.col(0);


			for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
			{
				
				tempM = H_m[rb_idx] * tempV.col(0);
				//tempIRC = tempM.transpose() * (tempM*tempM.transpose() + noise*eye + (0.1*randnum.u())*eye).inverse();
				tempIRC = tempM.transpose() * (tempM*tempM.transpose() + (0.1*randnum.u())*eye).inverse();
				temph = tempIRC*H_m[rb_idx];


				signal = temph*temph.transpose();



				//interferencePlusNoise = tempIRC*(noise*eye)*tempIRC.transpose();

				interferencePlusNoise = tempIRC*tempIRC.transpose();   //// noise  غ. noise  ڰ ʹ Գ

				

				CQI[rb_idx][abs(t) % cqi_history_length] = abs( real(signal(0, 0)) / real(interferencePlusNoise(0, 0)) );


				//// abs(밪)  KSIM µ,     쿡 ޺κп log2(1+cqi) Ҽ ͼ  ȵ

				/*
				cout << "tempV = " << tempV << endl;
				cout << "tempM = " << tempM << endl;
				cout << "tempIRC = " << tempIRC << endl;
				cout << "temph = " << temph << endl;
				cout << "noise = " << noise << endl;
				cout << "eye = " << eye << endl;
				cout << "tempIRC.transpose() = " << tempIRC.transpose() << endl;
				cout << "signal = " << signal << endl;
				cout << "interferencePlusNoise = " << interferencePlusNoise << endl;
				getchar();


				cout << "CQI = " << CQI[rb_idx][abs(t) % cqi_history_length] << endl;
				getchar();
				*/
			}

			double SUM = 0.;

			for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
			{
				SUM += log2(1.0 + CQI[rb_idx][abs(t) % cqi_history_length]);

				/*
				cout << "rb_idx = " << rb_idx << endl;
				cout << "CQI[rb_idx][abs(t) % cqi_history_length] = " << CQI[rb_idx][abs(t) % cqi_history_length] << endl;
				cout << "SUM = " << SUM << endl;
				getchar();
				*/
			}


			
			downlinkESINR = pow(2., SUM / num_rbs_in_1ms) - 1. + downlinkHARQeSINR;
			downlinkESINRdB = 10.0 * log10(downlinkESINR);
			downlinkspectralefficiency = GetSpectralEfficiency(downlinkESINRdB, ue_MCS);

			/*
			cout << "SUM / num_rbs_in_1ms = " << SUM / num_rbs_in_1ms << endl;
			cout << "downlinkHARQeSINR = " << downlinkHARQeSINR << endl;
			cout << "downlinkESINR = "<< downlinkESINR << endl;
			cout << "downlinkESINRdB = "<< downlinkESINRdB << endl;
			cout << "downlinkspectralefficiency = "<< downlinkspectralefficiency << endl;
			getchar();
			*/
		}

		else
		{

		}

		/*
		else   //// Use PMI
		{
			for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
			{

				//cmat M, A, B, C, w, u;

				MatrixXcd A;
				MatrixXcd B;
				VectorXcd w;
				Matrix2cd eye = Matrix2cd::Identity();

				double sinr_estimate;

				double geometry = link[_self_idx].str_signal - link[_self_idx].interference; // ̰Ŵ dB // ̰ (link[_self_idx].UE_PATH_LOSS / link[_self_idx].UE_ICI) ̰Ŷ  (linear)

				double linear_signal = dBm2linear(link[_self_idx].str_signal);
				double linear_interference = dBm2linear(link[_self_idx].interference);

				linear_interference = linear_interference + noise;

				
				w = code_word[PMI[rbs_idx][t%cqi_history_length]].conjugate();   // 4x1 vector
				A = linear_signal * H_m[rbs_idx] * w;   // 2x4 4x1
				B = linear_interference * eye;   /// 2x2 indentity
				
				// A norm  vector norm ,  ؼ ؼ Ʈ
				// B norm  matrix 2-norm, identity ׳ linear interference̶  


				///sinr_estimate = A.norm() / B.norm();
				sinr_estimate = A.norm() / linear_interference;

				//cout << sinr_estimate << endl;
				//  sinr_estimate = dB2linear(linear2dB( sinr_estimate ) + CQI_offset);


				CQI[rbs_idx][abs(t) % cqi_history_length] = sinr_estimate;
			}
		}
		*/
	}
}



double MS::GetSpectralEfficiency(double sinr_estimate, int &MCS)
{
	///


	if (sinr_estimate < -7.8474) { MCS = -1; return 0.0; }
	else if (sinr_estimate >= -7.8474 && sinr_estimate < -6.2369) { MCS = 0; return  0.1523; }
	else if (sinr_estimate >= -6.2369 && sinr_estimate < -4.3591) { MCS = 1; return 0.2344; }
	else if (sinr_estimate >= -4.3591 && sinr_estimate < -1.9319) { MCS = 2; return 0.3770; }
	else if (sinr_estimate >= -1.9319 && sinr_estimate < 0.1509) { MCS = 3; return 0.6016; }
	else if (sinr_estimate >= 0.1509 && sinr_estimate < 1.9976) { MCS = 4; return 0.8770; }
	else if (sinr_estimate >= 1.9976 && sinr_estimate < 4.7278) { MCS = 5; return 1.1758; }
	else if (sinr_estimate >= 4.7278 && sinr_estimate < 6.2231) { MCS = 6; return 1.4766; }
	else if (sinr_estimate >= 6.2231 && sinr_estimate < 8.0591) { MCS = 7; return 1.9141; }
	else if (sinr_estimate >= 8.0591 && sinr_estimate < 9.8585) { MCS = 8; return 2.4063; }
	else if (sinr_estimate >= 9.8585 && sinr_estimate < 11.8432) { MCS = 9; return 2.7305; }
	else if (sinr_estimate >= 11.8432 && sinr_estimate < 13.4893) { MCS = 10; return 3.3223; }
	else if (sinr_estimate >= 13.4893 && sinr_estimate < 15.3598) { MCS = 11; return 3.9023; }
	else if (sinr_estimate >= 15.3598 && sinr_estimate < 17.4435) { MCS = 12; return 4.5234; }
	else if (sinr_estimate >= 17.4435 && sinr_estimate < 19.2155) { MCS = 13; return 5.1152; }
	else if (sinr_estimate >= 19.2155) { MCS = 14; return 5.5547; }
	else
	{
		MCS = -1; return -1;
	}

	
}

void MS::Received_SINR(void)
{
	double noise = pow(10, (-174.0 / 10.0)) * (bandwidth);  // bandwidth = Hz
	MatrixXcd tempRI, tempRH, tempRHr, tempM, tempIRC, temph, eye, signal, interferencePlusNoise;
	tempRI = MatrixXcd::Zero(NUM_RX, NUM_RX);
	tempRHr = MatrixXcd::Zero(NUM_RX, NUM_RX);
	tempRH = MatrixXcd::Zero(NUM_TX, NUM_TX);
	eye = MatrixXcd::Identity(NUM_RX, NUM_RX);

	VectorXd FrequencySINR(num_rbs_in_1ms);

	// For SVD
	MatrixXcd tempV;


	for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
	{
		tempRHr = tempRHr + (H_m[rb_idx] * H_m[rb_idx].transpose()) / num_rbs_in_1ms;   /// tempRHr = num_RX x num_RX
		tempRH = tempRH + (H_m[rb_idx].transpose() * H_m[rb_idx]) / num_rbs_in_1ms;   /// tempRH  = num_TX x num_TX
	}

	JacobiSVD <MatrixXcd> svd(tempRH, ComputeFullU | ComputeFullV);    ////// U = RX x RX, V = TX x TX

	/*
	cout << "Its singular values are:" << endl << svd.singularValues() << endl;
	cout << "Its left singular vectors are the columns of the thin U matrix:" << endl << svd.matrixU() << endl;
	cout << "Its right singular vectors are the columns of the thin V matrix:" << endl << svd.matrixV() << endl;
	Vector3f rhs(1, 0, 0);
	cout << "Now consider this rhs vector:" << endl << rhs << endl;
	cout << "A least-squares solution of m*x = rhs is:" << endl << svd.solve(rhs) << endl;
	*/


	tempV = svd.matrixV();

	digitalPrecodingMatrix = tempV.col(0);


	for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
	{

		tempM = H_m[rb_idx] * tempV.col(0);
		//tempIRC = tempM.transpose() * (tempM*tempM.transpose() + noise*eye + (0.1*randnum.u())*eye).inverse();
		tempIRC = tempM.transpose() * (tempM*tempM.transpose() + (0.1*randnum.u())*eye).inverse();
		temph = tempIRC*H_m[rb_idx];


		signal = temph*temph.transpose();



		//interferencePlusNoise = tempIRC*(noise*eye)*tempIRC.transpose();

		interferencePlusNoise = tempIRC*tempIRC.transpose();   //// noise  غ. noise  ڰ ʹ Գ



		FrequencySINR(rb_idx) = abs(real(signal(0, 0)) / real(interferencePlusNoise(0, 0)));


		//// abs(밪)  KSIM µ,     쿡 ޺κп log2(1+cqi) Ҽ ͼ  ȵ

		/*
		cout << "tempV = " << tempV << endl;
		cout << "tempM = " << tempM << endl;
		cout << "tempIRC = " << tempIRC << endl;
		cout << "temph = " << temph << endl;
		cout << "noise = " << noise << endl;
		cout << "eye = " << eye << endl;
		cout << "tempIRC.transpose() = " << tempIRC.transpose() << endl;
		cout << "signal = " << signal << endl;
		cout << "interferencePlusNoise = " << interferencePlusNoise << endl;
		getchar();


		cout << "CQI = " << FrequencySINR(rb_idx) << endl;
		getchar();
		*/
	}

	double SUM = 0.;

	for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
	{
		SUM += log2(1.0 + FrequencySINR(rb_idx));

		/*
		cout << "rb_idx = " << rb_idx << endl;
		cout << "FrequencySINR(rb_idx) = " << FrequencySINR(rb_idx) << endl;
		cout << "SUM = " << SUM << endl;
		getchar();
		*/
	}

	downlinkESINR = pow(2., SUM / num_rbs_in_1ms) - 1. + downlinkHARQeSINR;
	downlinkESINRdB = 10.0 * log10(downlinkESINR);

	/*
	cout << downlinkESINR << endl;
	cout << downlinkESINRdB << endl;
	cout << downlinkspectralefficiency << endl;
	getchar();
	*/
}




void MS::Fourier_Transform_of_Channel(int ms_idx)
{
	if (CH_CAL == 1)   //// runtime calc
	{
		int self_idx = ms_idx;
		int A = 0;

		if (TYPE == 11)
		{
			A = link[self_idx]._sector_in_control;
		}
		else
		{
			A = (int)(link[self_idx].adj_sector[0] / 3);
		}

		for (int r_idx = 0; r_idx < NUM_RX; r_idx++) // Received    Channel
		{
			for (int t_idx = 0; t_idx < NUM_TX; t_idx++) // Transmitted Channel
			{
				int pointer = (int(DFT_SIZE) - (num_rbs_in_1ms*num_freq_per_rbs)) / 2;   ////  (1024 - 50*12) / 2 = 212
				double sum_for_normal = 0;
				for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++) // Subcarrier idx
				{
					int freq_idx = (pointer + (num_freq_per_rbs / 2) + (rbs_idx*num_freq_per_rbs)) * 15000; //// 15000 = 15kHz subcarrier   ///// (212 + (12/2) + (rbs_idx * 12)) * 15k
					for (int path_idx = 0; path_idx < channel[A][self_idx].NUM_PATH_for_channelcoeff; path_idx++)
					{
						complex <double> dft_exp(0.0, -2 * pi*freq_idx*(channel[A][self_idx].delay[path_idx]));
						H_m[rbs_idx](r_idx, t_idx) += channel[A][self_idx].CHIR[t_idx][r_idx][path_idx] * exp(dft_exp);

						/*
						////////////////////////// mu_offset_ZOD Ⱚ° ////////
						if (self_idx == 4)
						{
							cout << "channel[A][self_idx].NUM_PATH_for_channelcoeff = " << channel[A][self_idx].NUM_PATH_for_channelcoeff << endl;
							cout << "path_idx = " << path_idx << endl;
							cout << "A = " << A << endl;
							cout << "channel[A][self_idx].CHIR[t_idx][r_idx][path_idx] = " << channel[A][self_idx].CHIR[t_idx][r_idx][path_idx] << endl;
							getchar();
						}
						*/
					}
					sum_for_normal += pow(abs(H_m[rbs_idx](r_idx, t_idx)), 2.);
				}

				sum_for_normal = sum_for_normal / (double)num_rbs_in_1ms;

				for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++) // Subcarrier idx
				{
					H_m[rbs_idx](r_idx, t_idx) = H_m[rbs_idx](r_idx, t_idx) / sqrt(sum_for_normal);
				}
			}
		}
	}
}






void MS::Declare_ch_matrix(void)
{


	if (H_m == NULL)
	{
		H_m = new MatrixXcd[num_rbs_in_1ms];
	}



	for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
	{
		MatrixXcd temp(NUM_RX, NUM_TX);

		for (int r_idx = 0; r_idx < NUM_RX; r_idx++)
		{
			for (int t_idx = 0; t_idx < NUM_TX; t_idx++)
			{
				temp(r_idx, t_idx) = 0.;
			}
		}
		H_m[freq_idx] = temp;
	}

	//// H_m[num_rbs_in_1ms](NUM_RX,NUM_TX)


	if (HI_m == NULL)
	{
		HI_m = new MatrixXcd *[num_compute_coef - 1];
		for (int i = 0; i < num_compute_coef - 1; i++)
		{
			HI_m[i] = new MatrixXcd[num_rbs_in_1ms];
		}

	}


	for (int i = 0; i < num_compute_coef - 1; i++)
	{
		for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
		{
			MatrixXcd temp(NUM_RX, NUM_TX);

			for (int r_idx = 0; r_idx < NUM_RX; r_idx++)
			{
				for (int t_idx = 0; t_idx < NUM_TX; t_idx++)
				{
					temp(r_idx, t_idx) = 0.;
				}
			}
			HI_m[i][freq_idx] = temp;
		}
	}

	//// HI_m[num_compute_coef - 1][num_rbs_in_1ms](NUM_RX,NUM_TX)

	//cout << "HI_m = " << endl << HI_m[0][0] << endl;

	//cout << "HI_m = " << endl << HI_m[0][1] << endl;

}


/*
double Get_distance(LOCATION a, LOCATION b)
{
	return sqrt((a.x - b.x)*(a.x - b.x) + (a.y - b.y)*(a.y - b.y));
}
*/

////////////////////////////////////////////////// PMI  ̸ϴ° ߸ ڵ ///////////////////////////////////////////////////////////////////////////
/*
void MS::Precalculate_PMI(int _ms_idx, int _time)
{
	_self_idx = _ms_idx;
	int time_idx = _time;

	//clock_t begin, end;
	//begin = clock();

	Declare_ch_matrix_precalc();


	Fourier_Transform_of_Channel_precalc(_self_idx, time_idx);
	//begin = clock();

	Quantization_of_Ch_precalc(time_idx);
	//begin = clock();

	//CQI_Update();
	//end = clock();
	//

	//cout << "TIME = " << endl << (double)(end - begin) / CLOCKS_PER_SEC << endl;


}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void MS::Declare_ch_matrix_precalc(void)
{


MatrixXcd *Hm[20000][2][2];

for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
{
for (int los_idx = 0; los_idx < 2; los_idx++)
{
for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
{
Hm[time_idx][los_idx][indoor_idx] = new MatrixXcd[num_rbs_in_1ms];
}
}
}






for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
{
MatrixXcd temp(NUM_RX, NUM_TX);

for (int r_idx = 0; r_idx < NUM_RX; r_idx++)
{
for (int t_idx = 0; t_idx < NUM_TX; t_idx++)
{
temp(r_idx, t_idx) = 0.;
}
}
H_m[freq_idx] = temp;
}

//// H_m[num_rbs_in_1ms](NUM_RX,NUM_TX)


if (HI_m == NULL)
{
HI_m = new MatrixXcd *[num_compute_coef - 1];
for (int i = 0; i < num_compute_coef - 1; i++)
{
HI_m[i] = new MatrixXcd[num_rbs_in_1ms];
}

}


for (int i = 0; i < num_compute_coef - 1; i++)
{
for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
{
MatrixXcd temp(NUM_RX, NUM_TX);

for (int r_idx = 0; r_idx < NUM_RX; r_idx++)
{
for (int t_idx = 0; t_idx < NUM_TX; t_idx++)
{
temp(r_idx, t_idx) = 0.;
}
}
HI_m[i][freq_idx] = temp;
}
}

//// HI_m[num_compute_coef - 1][num_rbs_in_1ms](NUM_RX,NUM_TX)

//cout << "HI_m = " << endl << HI_m[0][0] << endl;

//cout << "HI_m = " << endl << HI_m[0][1] << endl;

}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void MS::Fourier_Transform_of_Channel_precalc(int ms_idx, int _time)
{

int self_idx = ms_idx;
int A = 0;

if (TYPE == 1)
{
A = link[self_idx]._sector_in_control;
}
else
{
A = link[self_idx].adj_sector[0] / 3;
}




for (int r_idx = 0; r_idx < NUM_RX; r_idx++) // Received    Channel
{
for (int t_idx = 0; t_idx < NUM_TX; t_idx++) // Transmitted Channel
{
int pointer = (int(DFT_SIZE) - (num_rbs_in_1ms*num_freq_per_rbs)) / 2;
double sum_for_normal = 0;

for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++) // Subcarrier idx
{
int freq_idx = (pointer + (num_freq_per_rbs / 2) + (rbs_idx*num_freq_per_rbs)) * 15000;

for (int path_idx = 0; path_idx < channel[A][self_idx].num_path; path_idx++)
{
/// num_path ׳ MAX num cluster
/// delay LOS / NLOS ؼ ϵ ϰ channel[0][0].delay_LOS[path_idx] ̷

complex <double> dft_exp(0.0, -2 * pi*freq_idx*(channel[A][self_idx].delay[path_idx]));




H_m[rbs_idx](r_idx, t_idx) += CHIR_precalc[_time][0][0][0][t_idx][r_idx][path_idx] * exp(dft_exp);

}

sum_for_normal += pow(abs(H_m[rbs_idx](r_idx, t_idx)), 2.);
}

sum_for_normal = sum_for_normal / (double)num_rbs_in_1ms;

for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++) // Subcarrier idx
{
H_m[rbs_idx](r_idx, t_idx) = H_m[rbs_idx](r_idx, t_idx) / sqrt(sum_for_normal);
}
}
}







}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void MS::Quantization_of_Ch_precalc(int _time)
{

for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
{

Matrix4cd Ohm;

double min = 500;

for (int cw_idx = 0; cw_idx < 16; cw_idx++)
{

Ohm = WW[cw_idx] * (H_m[rbs_idx].adjoint() * H_m[rbs_idx]) * WW[cw_idx].adjoint(); //Ohm = W HH H WH // 4x4 matrix

Matrix4cd UppOhm;
Matrix4cd LowOhm;

UppOhm = Ohm.triangularView<StrictlyUpper>(); //diagonal   triangle matrix
LowOhm = Ohm.triangularView<StrictlyLower>(); //diagonal  Ʒ triangle matrix

double temp = UppOhm.norm() + LowOhm.norm(); // minimizes off diagonal entries of Ohm

if (temp < min)
{
min = temp;
PMI[rbs_idx][_time % cqi_history_length] = cw_idx;
}
}


//PMI[rbs_idx][t%cqi_history_length] = rand() % 16;

//cout << "PMI = " << PMI[rbs_idx][t%cqi_history_length] << endl;

}

}
*/